from __future__ import annotations

from datetime import datetime
from enum import Enum
from functools import partial, wraps
from itertools import takewhile
import json
from pathlib import Path
import re
from typing import TYPE_CHECKING
from typing import Any, Callable, ClassVar, Dict, List, NamedTuple, Optional, Set, Tuple, Union

from loguru import logger
from yarl import URL

from . import exceptions as E
from .models import Pkg, PkgOptions
from .utils import ManagerAttrAccessMixin, gather, run_in_thread, slugify, bbegone, is_not_stale

try:
    from functools import singledispatchmethod      # type: ignore
except ImportError:
    from singledispatchmethod import singledispatchmethod

if TYPE_CHECKING:
    from .manager import Manager


_sentinel = object()

async_is_not_stale = run_in_thread(is_not_stale)
async_read = partial(run_in_thread(Path.read_text), encoding='utf-8')
async_write = partial(run_in_thread(Path.write_text), encoding='utf-8')


class Defn(NamedTuple):

    source: str
    name: str

    def __str__(self) -> str:
        return ':'.join(self)


class Strategies(str, Enum):

    default = 'default'
    latest = 'latest'


def validate_strategy(method: Callable) -> Callable:
    @wraps(method)
    async def wrapper(self, defns: Any, strategy_value: str, **kwargs: Any) -> Any:
        strategy = Strategies.__members__.get(strategy_value)
        if strategy in self.strategies:
            return await method(self, defns, strategy, **kwargs)
        raise E.PkgStrategyUnsupported(strategy_value)

    return wrapper


class _FileCacheMixin:

    async def _cache_json_response(self: Any, url: str, *args: Any) -> Any:
        from hashlib import md5

        filename = md5(url.encode()).hexdigest()
        path = self.config.temp_dir / f'{filename}.json'
        if await async_is_not_stale(path, *args):
            text = await async_read(path)
        else:
            async with self.web_client.get(url) as response:
                text = await response.text()
            await async_write(path, text)

        return json.loads(text)


class Resolver(ManagerAttrAccessMixin):

    source: ClassVar[str]
    name: ClassVar[str]
    strategies: ClassVar[Set[Strategies]]

    def __init__(self, *, manager: Manager) -> None:
        self.manager = manager

    @classmethod
    def decompose_url(cls, uri: str) -> Optional[Tuple[str, str]]:
        raise NotImplementedError

    async def synchronise(self) -> Resolver:
        return self

    async def resolve(self, defns: List[Defn], strategy: Strategies) -> List[Pkg]:
        raise NotImplementedError


class CurseResolver(Resolver, _FileCacheMixin):

    source = 'curse'
    name = 'CurseForge'
    strategies = {Strategies.default, Strategies.latest}

    # https://twitchappapi.docs.apiary.io/
    addon_api_url = 'https://addons-ecs.forgesvc.net/api/v2/addon'
    slugs_url = ('https://raw.githubusercontent.com/layday/instascrape/data/'
                 'curseforge-slugs.json')   # v1
    folders_url = ('https://raw.githubusercontent.com/layday/instascrape/data/'
                   'curseforge-folders.json')   # v1

    _slugs: Optional[Dict[str, str]] = None

    @classmethod
    def decompose_url(cls, uri: str) -> Optional[Tuple[str, str]]:
        url = URL(uri)
        if url.host == 'www.wowace.com' \
                and len(url.parts) > 2 \
                and url.parts[1] == 'projects':
            return (cls.source, url.parts[2].lower())
        elif url.host == 'www.curseforge.com' \
                and len(url.parts) > 3 \
                and url.parts[1:3] == ('wow', 'addons'):
            return (cls.source, url.parts[3].lower())

    async def synchronise(self) -> CurseResolver:
        if self._slugs is None:
            slugs = await self._cache_json_response(self.slugs_url, 8, 'hours')
            self._slugs = {k: str(v) for k, v in slugs.items()}
        return self

    @singledispatchmethod
    @validate_strategy
    async def resolve(self, defns: List[Defn], strategy: Strategies) -> List[Pkg]:
        assert self._slugs is not None, f'{self.source} has not been synchronised'

        ids_from_defns = [self._slugs.get(d.name, d.name) for d in defns]
        numeric_ids = list({i for i in ids_from_defns if i.isdigit()})
        async with self.web_client.post(self.addon_api_url, json=numeric_ids) as response:
            if response.status == 404:
                api_results = []       # type: ignore
            else:
                api_results = await response.json()

        results = {str(r['id']): r for r in api_results}
        return await gather(self.resolve(d, strategy, _metadata=results.get(i))
                            for d, i in zip(defns, ids_from_defns))

    @resolve.register
    @validate_strategy
    async def _(self, defn: Defn, strategy: Strategies, *, _metadata: Any = _sentinel) -> Pkg:
        if _metadata is _sentinel:
            pkg, = await self.resolve([defn], strategy)
            return pkg
        else:
            metadata = _metadata

        if not metadata:
            raise E.PkgNonexistent
        elif not metadata['latestFiles']:
            raise E.PkgFileUnavailable('no files available for download')

        flavor = 'wow_classic' if self.config.is_classic else 'wow_retail'
        files = (f for f in metadata['latestFiles']
                 if not f['isAlternate']    # nolib file if true
                 and f['gameVersionFlavor'] == flavor)
        if strategy is Strategies.default:
            # 1 = stable
            # 2 = beta
            # 3 = alpha
            files = (f for f in files if f['releaseType'] == 1)
        try:
            _, file = max((f['id'], f) for f in files)
        except ValueError:
            raise E.PkgFileUnavailable('no files meet criteria')

        return Pkg(origin=self.source,
                   id=metadata['id'],
                   slug=metadata['slug'],
                   name=metadata['name'],
                   description=metadata['summary'],
                   url=metadata['websiteUrl'],
                   file_id=file['id'],
                   download_url=file['downloadUrl'],
                   date_published=file['fileDate'],
                   version=file['displayName'],
                   options=PkgOptions(strategy=strategy.name))


class WowiResolver(Resolver, _FileCacheMixin):

    source = 'wowi'
    name = 'WoWInterface'
    strategies = {Strategies.default}

    # https://api.mmoui.com/v3/globalconfig.json
    list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    _files: Optional[Dict[str, dict]] = None

    @classmethod
    def decompose_url(cls, uri: str) -> Optional[Tuple[str, str]]:
        url = URL(uri)
        if url.host in {'wowinterface.com', 'www.wowinterface.com'} \
                and len(url.parts) == 3 \
                and url.parts[1] == 'downloads':
            if url.name == 'landing.php':
                id_ = url.query.get('fileid')
                if id_:
                    return (cls.source, id_)
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)(?:-(?P<name>[^\.]+))?',
                                 url.name)
                if match:
                    id_, slug = match.groups()
                    if slug:
                        id_ = slugify(f'{id_} {slug}')
                    return (cls.source, id_)

    async def synchronise(self) -> WowiResolver:
        if self._files is None:
            files = await self._cache_json_response(self.list_api_url, 3600)
            self._files = {i['UID']: i for i in files}
        return self

    @singledispatchmethod
    @validate_strategy
    async def resolve(self, defns: List[Defn], strategy: Strategies) -> List[Pkg]:
        assert self._files is not None, f'{self.source} has not been synchronised'

        ids_from_defns = [''.join(takewhile(str.isdigit, d.name)) for d in defns]
        numeric_ids = {i for i in ids_from_defns if i.isdigit()}
        url = self.details_api_url / f'{",".join(numeric_ids)}.json'
        async with self.web_client.get(url) as response:
            if response.status == 404:
                api_results = []       # type: ignore
            else:
                api_results = await response.json()

        results = {r['UID']: {**self._files[r['UID']], **r} for r in api_results}
        return await gather(self.resolve(d, strategy, _metadata=results.get(i))
                            for d, i in zip(defns, ids_from_defns))

    @resolve.register
    @validate_strategy
    async def _(self, defn: Defn, strategy: Strategies, *, _metadata: Any = _sentinel) -> Pkg:
        if _metadata is _sentinel:
            pkg, = await self.resolve([defn], strategy)
            return pkg
        else:
            metadata = _metadata

        if not metadata:
            raise E.PkgNonexistent
        if metadata['UIPending'] == '1':
            raise E.PkgFileUnavailable('new file awaiting approval')

        # The file is only assumed to be compatible with classic if 'WoW Classic'
        # is listed under compatibility, and incompatible with retail if
        # 'WoW Classic' is the sole element in the array.  I'm not sure
        # when 'UICompatibility' was added but it's very often not populated
        # for retail add-ons *shrugsies*
        compatibility = {e['name'] for e in metadata['UICompatibility'] or ()}
        if self.config.is_classic:
            if 'WoW Classic' not in compatibility:
                raise E.PkgFileUnavailable('file is not compatible with classic')
        elif compatibility and not compatibility - {'WoW Classic'}:
            raise E.PkgFileUnavailable('file is only compatible with classic')

        return Pkg(origin=self.source,
                   id=metadata['UID'],
                   slug=slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
                   name=metadata['UIName'],
                   description=bbegone(metadata['UIDescription']),
                   url=metadata['UIFileInfoURL'],
                   file_id=metadata['UIMD5'],
                   download_url=metadata['UIDownload'],
                   date_published=metadata['UIDate'],
                   version=metadata['UIVersion'],
                   options=PkgOptions(strategy=strategy.name))


class TukuiResolver(Resolver):

    source = 'tukui'
    name = 'Tukui'
    strategies = {Strategies.default}

    api_url = URL('https://www.tukui.org/api.php')

    @classmethod
    def decompose_url(cls, uri: str) -> Optional[Tuple[str, str]]:
        url = URL(uri)
        if url.host == 'www.tukui.org' \
                and url.path in {'/addons.php', '/classic-addons.php', '/download.php'}:
            id_or_slug = url.query.get('id') or url.query.get('ui')
            if id_or_slug:
                return (cls.source, id_or_slug)

    @singledispatchmethod
    @validate_strategy
    async def resolve(self, defns: List[Defn], strategy: Strategies) -> List[Pkg]:
        return await gather(self.resolve(d, strategy) for d in defns)

    @resolve.register
    @validate_strategy
    async def _(self, defn: Defn, strategy: Strategies) -> Pkg:
        id_, *_ = defn.name.partition('-')
        is_ui = id_ in {'elvui', 'tukui'}

        if self.config.is_classic:
            query = 'classic-addon'
        else:
            if is_ui:
                query = 'ui'
            else:
                query = 'addon'
        url = self.api_url.with_query({query: id_})
        async with self.web_client.get(url) as response:
            if not response.content_length:
                raise E.PkgNonexistent
            addon = await response.json(content_type='text/html')

        if is_ui:
            slug = id_
            date_published = datetime.fromisoformat(addon['lastupdate'])
        else:
            slug = slugify(f'{addon["id"]} {addon["name"]}')
            date_published = addon['lastupdate']

        return Pkg(origin=self.source,
                   id=addon['id'],
                   slug=slug,
                   name=addon['name'],
                   description=addon['small_desc'],
                   url=addon['web_url'],
                   file_id=addon['lastupdate'],
                   download_url=addon['url'],
                   date_published=date_published,
                   version=addon['version'],
                   options=PkgOptions(strategy=strategy.name))


class InstawowResolver(Resolver):

    source = 'instawow'
    name = 'instawow'
    strategies = {Strategies.default}

    _addons = {('0', 'weakauras-companion'),
               ('1', 'weakauras-companion-autoupdate')}

    @classmethod
    def decompose_url(cls, uri: str) -> None:
        return

    @singledispatchmethod
    @validate_strategy
    async def resolve(self, defns: List[Defn], strategy: Strategies) -> List[Pkg]:
        return await gather(self.resolve(d, strategy) for d in defns)

    @resolve.register
    @validate_strategy
    async def _(self, defn: Defn, strategy: Strategies) -> Pkg:
        try:
            id_, slug = next(p for p in self._addons if defn.name in p)
        except StopIteration:
            raise E.PkgNonexistent

        from .wa_updater import WaCompanionBuilder

        builder = WaCompanionBuilder(self.manager)
        if id_ == '1':
            try:
                await builder.build()
            except ValueError as error:
                raise E.PkgFileUnavailable('account name not provided') from error

        return Pkg(origin=self.source,
                   id=id_,
                   slug=slug,
                   name='WeakAuras Companion',
                   description='A WeakAuras Companion clone.',
                   url='https://github.com/layday/instawow',
                   file_id=await run_in_thread(builder.checksum)(),
                   download_url=builder.file_out.as_uri(),
                   date_published=datetime.now(),
                   version='1.0.0',
                   options=PkgOptions(strategy=strategy.name))



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
import contextvars as cv
from functools import partial
from pathlib import Path, PurePath
import posixpath
from shutil import move as _move
from tempfile import NamedTemporaryFile, mkdtemp
from typing import *

from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from . import db_version
from . import exceptions as E
from .models import ModelBase, Pkg, PkgFolder, should_migrate
from .resolvers import Defn, CurseResolver, WowiResolver, TukuiResolver, InstawowResolver
from .utils import (bucketise, cached_property, dict_merge, gather,
                    iter_or_repeat, make_progress_bar, run_in_thread)

if TYPE_CHECKING:
    from types import SimpleNamespace
    import aiohttp
    from .config import Config


USER_AGENT = 'instawow (https://github.com/layday/instawow)'

_web_client: cv.ContextVar[aiohttp.ClientSession] = cv.ContextVar('_web_client')

AsyncNamedTemporaryFile = partial(run_in_thread(NamedTemporaryFile), prefix='instawow-')
async_mkdtemp = run_in_thread(mkdtemp)
async_move = run_in_thread(_move)


@asynccontextmanager
async def open_temp_writer() -> AsyncGenerator[Tuple[Path, Callable], None]:
    fh = await AsyncNamedTemporaryFile(delete=False, suffix='.zip')
    try:
        yield (Path(fh.name), run_in_thread(fh.write))
    finally:
        await run_in_thread(fh.close)()


async def new_temp_dir(*args: Any, **kwargs: Any) -> PurePath:
    return PurePath(await async_mkdtemp(*args, **kwargs))


async def move(paths: Iterable[Path], dest: PurePath) -> None:
    for path in paths:
        await async_move(str(path), dest)


# macOS 'resource forks' are sometimes included in download zips - these
# violate our one add-on per folder contract-thing, and serve absolutely
# no purpose.  They won't be extracted and they won't be added to the database
_zip_excludes = {'__MACOSX'}


def _find_base_dirs(names):
    return {n for n in (posixpath.dirname(n) for n in names)
            if n and posixpath.sep not in n} - _zip_excludes


def _should_extract(base_dirs):
    def is_member(name):
        head, sep, _ = name.partition(posixpath.sep)
        return sep and head in base_dirs

    return is_member


def Archive(path: PurePath, delete_after: bool) -> Callable:
    from zipfile import ZipFile

    @asynccontextmanager
    async def enter() -> AsyncGenerator[Tuple[List[str], Callable], None]:
        zip_file = await run_in_thread(ZipFile)(path)
        names = zip_file.namelist()
        base_dirs = _find_base_dirs(names)

        def extract(parent: Path) -> None:
            conflicts = base_dirs & {f.name for f in parent.iterdir()}
            if conflicts:
                raise E.PkgConflictsWithUncontrolled(conflicts)
            else:
                members = filter(_should_extract(base_dirs), names)
                zip_file.extractall(parent, members=members)

        def exit_() -> None:
            zip_file.close()
            if delete_after:
                path.unlink()   # type: ignore

        try:
            yield (sorted(base_dirs), run_in_thread(extract))
        finally:
            await run_in_thread(exit_)()

    return enter


async def download_archive(pkg: Pkg, *, chunk_size: int = 4096) -> Callable:
    url = pkg.download_url
    if url.startswith('file://'):
        from urllib.parse import unquote

        path = PurePath(unquote(url[7:]))
        return Archive(path, delete_after=False)
    else:
        async with (_web_client.get()
                    .get(url, trace_request_ctx={'show_progress': True})) as response, \
                   open_temp_writer() as (path, write):
            async for chunk in response.content.iter_chunked(chunk_size):
                await write(chunk)
        return Archive(path, delete_after=True)


def prepare_db_session(config: Config) -> scoped_session:
    db_url = f"sqlite:///{config.config_dir / 'db.sqlite'}"
    engine = create_engine(db_url)
    ModelBase.metadata.create_all(engine)

    if should_migrate(engine, db_version):
        from alembic.autogenerate import compare_metadata
        from alembic.migration import MigrationContext
        from .migrations import make_config, stamp, upgrade

        with engine.begin() as conn:
            alembic_config = make_config(db_url)
            diff = compare_metadata(MigrationContext.configure(conn),
                                    ModelBase.metadata)
            if diff:
                logger.info(f'migrating database to {db_version}')
                upgrade(alembic_config, db_version)
            else:
                logger.info(f'stamping database with {db_version}')
                stamp(alembic_config, db_version)

    return scoped_session(sessionmaker(bind=engine))


async def init_web_client(**kwargs: Any) -> aiohttp.ClientSession:
    from aiohttp import ClientSession, TCPConnector

    kwargs = {'connector': TCPConnector(force_close=True, limit_per_host=10),
              'headers': {'User-Agent': USER_AGENT},
              'trust_env': True,
              **kwargs}
    return ClientSession(**kwargs)


class _DummyResolver:

    async def synchronise(*args: Any, **kwargs: Any) -> Type[_DummyResolver]:
        return _DummyResolver

    async def resolve(*args: Any, **kwargs: Any) -> NoReturn:
        raise E.PkgOriginInvalid


class _ResolverDict(dict):

    RESOLVERS = {CurseResolver, WowiResolver, TukuiResolver, InstawowResolver}

    def __init__(self, manager: Manager) -> None:
        super().__init__((r.source, r(manager=manager)) for r in self.RESOLVERS)

    def __missing__(self, key: Hashable) -> Type[_DummyResolver]:
        return _DummyResolver


def _error_wrapper(error: E.ManagerError) -> Callable:
    async def error_out() -> NoReturn:
        raise error
    return error_out


class Manager:

    def __init__(self, config: Config, db_session: scoped_session) -> None:
        self.config = config
        self.db_session = db_session
        self.resolvers = _ResolverDict(self)

    @property
    def web_client(self) -> aiohttp.ClientSession:
        return _web_client.get()

    @web_client.setter
    def web_client(self, value: aiohttp.ClientSession) -> None:
        _web_client.set(value)

    def get(self, defn: Tuple[str, str]) -> Pkg:
        "Retrieve a package from the database."
        source, name = defn
        return (self.db_session.query(Pkg)
                .filter(Pkg.origin == source, (Pkg.id == name) | (Pkg.slug == name))
                .first())

    async def resolve(self, defns: Sequence[Defn], strategy: str) -> Dict[Defn, Any]:
        "Resolve a sequence of URIs into packages."
        defns_by_source = bucketise(defns, key=lambda v: v.source)
        results = await gather([((await self.resolvers[s].synchronise())
                                 .resolve(list(b), strategy))
                                for s, b in defns_by_source.items()])
        matched = dict_merge(dict.fromkeys(defns),
                             {d: r
                              for ds, rs in zip(defns_by_source.values(), results)
                              for d, r in zip(ds, iter_or_repeat(rs))})
        return matched

    async def _install(self, pkg: Pkg, open_archive: Callable, replace: bool) -> E.PkgInstalled:
        async with open_archive() as (folders, extract):
            conflicts = (self.db_session.query(PkgFolder)
                         .filter(PkgFolder.name.in_(folders)).all())
            if conflicts:
                raise E.PkgConflictsWithInstalled([c.pkg for c in conflicts])
            if replace:
                temp_dir = await new_temp_dir(dir=self.config.temp_dir, prefix=f'{folders[0]}-')
                await move((self.config.addon_dir / f for f in folders), temp_dir)
            await extract(self.config.addon_dir)

        pkg.folders = [PkgFolder(name=f) for f in folders]
        self.db_session.add(pkg)
        self.db_session.commit()
        return E.PkgInstalled(pkg)

    async def prep_install(self, uris: Sequence[Defn], strategy: str, replace: bool) -> Dict[Defn, Coroutine]:
        "Retrieve packages to install."
        results = await self.resolve([u for u in uris if not self.get(u)], strategy)
        installables = {(u, r): download_archive(r) for u, r in results.items()
                        if isinstance(r, Pkg)}
        archives = await gather(installables.values())

        coros = {k: v() for k, v in
                 dict_merge(dict.fromkeys(uris, _error_wrapper(E.PkgAlreadyInstalled())),
                            {u: _error_wrapper(r) for u, r in results.items()},
                            {u: (lambda p=p, a=a: self._install(p, a, replace))
                             for (u, p), a in zip(installables, archives)}).items()}
        return coros

    async def _update(self, old_pkg: Pkg, new_pkg: Pkg, open_archive: Callable) -> E.PkgUpdated:
        async with open_archive() as (folders, extract):
            filter_ = (PkgFolder.name.in_(folders),
                       PkgFolder.pkg_origin != new_pkg.origin,
                       PkgFolder.pkg_id != new_pkg.id)
            conflicts = self.db_session.query(PkgFolder).filter(*filter_).first()
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts.pkg)

            temp_dir = await new_temp_dir(dir=self.config.temp_dir,
                                          prefix=f'{old_pkg.folders[0].name}-')
            await move((self.config.addon_dir / f.name for f in old_pkg.folders), temp_dir)
            await extract(self.config.addon_dir)

        new_pkg.folders = [PkgFolder(name=f) for f in folders]
        self.db_session.delete(old_pkg)
        self.db_session.add(new_pkg)
        self.db_session.commit()
        return E.PkgUpdated(old_pkg, new_pkg)

    async def prep_update(self, uris: Sequence[Defn]) -> Dict[Defn, Coroutine]:
        "Retrieve packages to update."
        candidates = {u: p for u, p in ((u, self.get(u)) for u in uris) if p}
        candidates_by_strategy = bucketise(candidates.items(),
                                           key=lambda v: v[1].options.strategy)
        results = dict_merge(*await gather(self.resolve([u for u, _ in c], s)
                                           for s, c in candidates_by_strategy.items()))
        installables = [(u, p) for u, p in results.items() if isinstance(p, Pkg)]
        updatables = {(u, p): download_archive(p) for u, p in installables
                      if p.file_id != candidates[u].file_id}
        archives = await gather(updatables.values())

        coros = {k: v() for k, v in
                 dict_merge(dict.fromkeys(uris, _error_wrapper(E.PkgNotInstalled())),
                            {u: _error_wrapper(r) for u, r in results.items()},
                            {u: _error_wrapper(E.PkgUpToDate()) for u, _ in installables},
                            {u: (lambda o=candidates[u], p=p, a=a: self._update(o, p, a))
                             for (u, p), a in zip(updatables, archives)}).items()}
        return coros

    async def _remove(self, pkg: Pkg) -> E.PkgRemoved:
        temp_dir = await new_temp_dir(dir=self.config.temp_dir,
                                      prefix=f'{pkg.folders[0].name}-')
        await move((self.config.addon_dir / f.name for f in pkg.folders), temp_dir)

        self.db_session.delete(pkg)
        self.db_session.commit()
        return E.PkgRemoved(pkg)

    async def prep_remove(self, uris: Sequence[Defn]) -> Dict[Defn, Coroutine]:
        "Prepare packages to remove."
        coros = {k: v() for k, v in
                 dict_merge(dict.fromkeys(uris, _error_wrapper(E.PkgNotInstalled())),
                            {u: (lambda p=p: self._remove(p))
                             for u, p in ((u, self.get(u)) for u in uris) if p}).items()}
        return coros


_tick_interval = .1
_tickers: cv.ContextVar[Set[asyncio.Task]] = cv.ContextVar('_tickers', default=set())


@asynccontextmanager
async def cancel_tickers() -> AsyncGenerator[None, None]:
    try:
        yield
    finally:
        for ticker in _tickers.get():
            ticker.cancel()


async def init_cli_web_client(*, manager: CliManager) -> aiohttp.ClientSession:
    "A web client that interfaces with the manager's progress bar."
    from cgi import parse_header
    from aiohttp import TraceConfig

    def extract_filename(params: aiohttp.TraceRequestEndParams) -> str:
        cd = params.response.headers.get('Content-Disposition', '')
        _, cd_params = parse_header(cd)
        filename = cd_params.get('filename') or params.response.url.name
        return filename

    async def do_on_request_end(session: aiohttp.ClientSession,
                                ctx: SimpleNamespace,
                                params: aiohttp.TraceRequestEndParams) -> None:
        # Requests don't have a context unless they
        # originate from ``download_archive``
        if ctx.trace_request_ctx is None:
            return

        bar = manager.bar(label=f'Downloading {extract_filename(params)}',
                          total=params.response.content_length)

        async def ticker(bar=bar, params=params) -> None:
            try:
                while not params.response.content._eof:
                    bar.current = params.response.content._cursor
                    await asyncio.sleep(_tick_interval)
            finally:
                bar.progress_bar.counters.remove(bar)

        tickers = _tickers.get()
        tickers.add(asyncio.create_task(ticker()))

    trace_config = TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()
    return await init_web_client(trace_configs=[trace_config])


async def _intercept(coro: Coroutine) -> E.ManagerResult:
    try:
        return await coro
    except E.ManagerError as error:
        return error
    except Exception as error:
        logger.exception('internal error')
        return E.InternalError(error=error)


class CliManager(Manager):

    def __init__(self, config: Config, db_session: scoped_session,
                 progress_bar_factory: Optional[Callable] = None) -> None:
        super().__init__(config, db_session)
        self.progress_bar_factory = progress_bar_factory or make_progress_bar
        self.loop = asyncio.new_event_loop()

    @logger.catch(reraise=True)
    def _process(self, prepper: Callable,
                 uris: Sequence[Defn], *args: Any, **kwargs: Any) -> List[E.ManagerResult]:
        async def do_process():
            async with cancel_tickers():
                coros = await prepper(list(uris), *args, **kwargs)
                return [await _intercept(c) for c in coros.values()]

        with self.bar:
            return self.run(do_process())

    def run(self, awaitable: Awaitable) -> Any:
        "Run ``awaitable`` inside an explicit context."
        async def do_run():
            async with (await init_cli_web_client(manager=self)) as self.web_client:
                return await awaitable

        runner = lambda: self.loop.run_until_complete(do_run())
        return cv.copy_context().run(runner)

    @cached_property
    def bar(self) -> Any:
        return self.progress_bar_factory()

    @property
    def install(self) -> Callable:
        return partial(self._process, self.prep_install)

    @property
    def update(self) -> Callable:
        return partial(self._process, self.prep_update)

    @property
    def remove(self) -> Callable:
        return partial(self._process, self.prep_remove)
